@@ -78,6 +78,7 @@ static char record_buffer[8192];
 static uint32 record_buffer_size;
 
 static uint8 packet_opcode;
+static size_t packet_len;
 static struct stream packet;
 
 void (*wave_out_play) (void);
@@ -116,6 +117,7 @@ rdpsnd_send_completion(uint16 tick, uint8 packet_index)
 	out_uint8(s, 0);
 	s_mark_end(s);
 	rdpsnd_send(s);
+	s_free(s);
 
 	DEBUG_SOUND(("RDPSND: -> RDPSND_COMPLETION(tick: %u, index: %u)\n",
 		     (unsigned) tick, (unsigned) packet_index));
@@ -148,10 +150,11 @@ rdpsnd_flush_record(void)
 			chunk_size = 1596;
 
 		s = rdpsnd_init_packet(RDPSND_REC_DATA, chunk_size);
-		out_uint8p(s, data, chunk_size);
+		out_uint8a(s, data, chunk_size);
 
 		s_mark_end(s);
 		rdpsnd_send(s);
+		s_free(s);
 
 		data = data + chunk_size;
 		record_buffer_size -= chunk_size;
@@ -346,6 +349,7 @@ rdpsnd_process_negotiate(STREAM in)
 	DEBUG_SOUND(("RDPSND: -> RDPSND_NEGOTIATE(formats: %d)\n", (int) format_count));
 
 	rdpsnd_send(out);
+	s_free(out);
 
 	rdpsnd_negotiated = True;
 }
@@ -359,7 +363,7 @@ rdpsnd_process_ping(STREAM in)
 
 	if (!s_check_rem(in, 4))
 	{
-		rdp_protocol_error("rdpsnd_process_training(), consume of training data from stream would overrun", &packet);
+		rdp_protocol_error("consume of training data from stream would overrun", &packet);
 	}
 
 	in_uint16_le(in, tick);
@@ -371,6 +375,7 @@ rdpsnd_process_ping(STREAM in)
 	out_uint16_le(out, 0);
 	s_mark_end(out);
 	rdpsnd_send(out);
+	s_free(out);
 
 	DEBUG_SOUND(("RDPSND: -> (tick: 0x%04x)\n", (unsigned) tick));
 }
@@ -470,6 +475,7 @@ rdpsnd_process_rec_negotiate(STREAM in)
 	DEBUG_SOUND(("RDPSND: -> RDPSND_REC_NEGOTIATE(formats: %d)\n", (int) rec_format_count));
 
 	rdpsnd_send(out);
+	s_free(out);
 
 	rdpsnd_rec_negotiated = True;
 }
@@ -478,8 +484,11 @@ static void
 rdpsnd_process_packet(uint8 opcode, STREAM s)
 {
 	uint16 vol_left, vol_right;
-	static uint16 tick, format;
-	static uint8 packet_index;
+
+	uint16 tick, format;
+	uint8 packet_index;
+	unsigned int size;
+	unsigned char *data;
 
 	switch (opcode)
 	{
@@ -524,9 +533,12 @@ rdpsnd_process_packet(uint8 opcode, STREAM s)
 				current_format = format;
 			}
 
-			rdpsnd_queue_write(rdpsnd_dsp_process
-					   (s->p, s->end - s->p, current_driver,
-					    &formats[current_format]), tick, packet_index);
+			size = s_remaining(s);
+			in_uint8p(s, data, size);
+			rdpsnd_queue_write(rdpsnd_dsp_process(data, size,
+							      current_driver,
+							      &formats[current_format]),
+					   tick, packet_index);
 			return;
 			break;
 		case RDPSND_CLOSE:
@@ -602,39 +614,35 @@ rdpsnd_process_packet(uint8 opcode, STREAM s)
 static void
 rdpsnd_process(STREAM s)
 {
-	uint16 len;
-
 	while (!s_check_end(s))
 	{
 		/* New packet */
-		if (packet.size == 0)
+		if (packet_len == 0)
 		{
-			if ((s->end - s->p) < 4)
+			if (!s_check_rem(s, 4))
 			{
 				error("RDPSND: Split at packet header. Things will go south from here...\n");
 				return;
 			}
 			in_uint8(s, packet_opcode);
 			in_uint8s(s, 1);	/* Padding */
-			in_uint16_le(s, len);
+			in_uint16_le(s, packet_len);
 
 			DEBUG_SOUND(("RDPSND: == Opcode %x Length: %d ==\n",
 				     (int) packet_opcode, (int) len));
-
-			packet.p = packet.data;
-			packet.end = packet.data + len;
-			packet.size = len;
 		}
 		else
 		{
-			len = MIN(s->end - s->p, packet.end - packet.p);
+			uint16 len;
+
+			len = MIN(s_remaining(s), packet_len - s_length(&packet));
 
 			/* Microsoft's server is so broken it's not even funny... */
 			if (packet_opcode == RDPSND_WRITE)
 			{
-				if ((packet.p - packet.data) < 12)
-					len = MIN(len, 12 - (packet.p - packet.data));
-				else if ((packet.p - packet.data) == 12)
+				if (s_length(&packet) < 12)
+					len = MIN(len, 12 - s_length(&packet));
+				else if (s_length(&packet) == 12)
 				{
 					DEBUG_SOUND(("RDPSND: Eating 4 bytes of %d bytes...\n",
 						     len));
@@ -643,16 +651,18 @@ rdpsnd_process(STREAM s)
 				}
 			}
 
-			in_uint8a(s, packet.p, len);
-			packet.p += len;
+			in_uint8stream(s, &packet, len);
+			/* Always end it so s_length() works */
+			s_mark_end(&packet);
 		}
 
 		/* Packet fully assembled */
-		if (packet.p == packet.end)
+		if (s_length(&packet) == packet_len)
 		{
-			packet.p = packet.data;
+			s_seek(&packet, 0);
 			rdpsnd_process_packet(packet_opcode, &packet);
-			packet.size = 0;
+			packet_len = 0;
+			s_reset(&packet);
 		}
 	}
 }
@@ -673,15 +683,11 @@ rdpsnddbg_process(STREAM s)
 	static char *rest = NULL;
 	char *buf;
 
-	if (!s_check(s))
-	{
-		rdp_protocol_error("rdpsnddbg_process(), stream is in unstable state", s);
-	}
-
-	pkglen = s->end - s->p;
+	pkglen = s_remaining(s);
 	/* str_handle_lines requires null terminated strings */
 	buf = (char *) xmalloc(pkglen + 1);
-	STRNCPY(buf, (char *) s->p, pkglen + 1);
+	in_uint8a(s, buf, pkglen);
+	buf[pkglen] = '\0';
 
 	str_handle_lines(buf, &rest, rdpsnddbg_line_handler, NULL);
 
@@ -732,9 +738,7 @@ rdpsnd_init(char *optarg)
 
 	drivers = NULL;
 
-	packet.data = (uint8 *) xmalloc(65536);
-	packet.p = packet.end = packet.data;
-	packet.size = 0;
+	s_realloc(&packet, 65536);
 
 	rdpsnd_channel =
 		channel_register("rdpsnd", CHANNEL_OPTION_INITIALIZED | CHANNEL_OPTION_ENCRYPT_RDP,
@@ -865,7 +869,7 @@ rdpsnd_queue_write(STREAM s, uint16 tick, uint8 index)
 
 	queue_hi = next_hi;
 
-	packet->s = *s;
+	packet->s = s;
 	packet->tick = tick;
 	packet->index = index;
 
@@ -899,7 +903,7 @@ rdpsnd_queue_clear(void)
 	while (queue_pending != queue_hi)
 	{
 		packet = &packet_queue[queue_pending];
-		xfree(packet->s.data);
+		s_free(packet->s);
 		queue_pending = (queue_pending + 1) % MAX_QUEUE;
 	}
 
@@ -964,7 +968,7 @@ rdpsnd_queue_complete_pending(void)
 			(packet->completion_tv.tv_usec - packet->arrive_tv.tv_usec);
 		elapsed /= 1000;
 
-		xfree(packet->s.data);
+		s_free(packet->s);
 		rdpsnd_send_completion((packet->tick + elapsed) % 65536, packet->index);
 		queue_pending = (queue_pending + 1) % MAX_QUEUE;
 	}
